/*****************************************************************************/
/*** Main driver of the code to calculate the biased angular power spectra  **/
/*** of the galaxy distribution, Tl (really, ratios to fiducial Cl), given  **/
/*** input of calibration errors in dN/N or mag with arbitrary spatial dep. **/
/*************************************************************************************************/
/***                   RIGHT NOW TAKES SOME INPUT FROM KEYBOARD                                 **/
/***  TO CHANGE THE INPUT TO ALL-FROM-FILE, CHANGE THE VARIABLE 'ALL_FILE_INPUT' IN declare.h   **/
/*************************************************************************************************/
#include<omp.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<math.h>
#include<time.h>

#include"declare.h"
#include"nrutil.h"
#include"nr.h"
#include"wigner.h"

#define MAX_LENGTH 130  // maximum length of a single *line*, with comment, in all_file_input.dat
    
int main(int argc, char **argv) 
{
    int i, l, iz, jz, il =0, il2, l1, m1, alpha, N_L2;
    int ALL_FILE_INPUT, LMAX_SPREAD_CVAR , CHOICE, CAL_OR_MAG;
    double CALIB_OR_MAG_ERR ;
    double *mag_to_calib_fac, *sz_arr;
    double **Cl_matrix, **Tl_matrix, **Cl_all_ell, *Cl_calib;
    double ***c_arr;
    char term, mystring[MAX_LENGTH], filename[200];
    FILE *ifp, *ifp_all_file_input;


    if (argc < 2) // input will be from keyboard
    {
        printf("NO INPUT FILE SELECTED. Now answer these questions: \n");      
        ALL_FILE_INPUT =0;
    }
    
    if (argc>1) // input is automatically from the file specified
    {
        strcpy(filename,argv[1]);
        ALL_FILE_INPUT=1;
        
        ifp_all_file_input = fopen(filename, "r");
        if (ifp_all_file_input == NULL) {
            fprintf(stderr, "Error: failed to open file %s\n", filename);
            printf("Exiting...\n");
            exit(1);
        }
    }    
    
    /****************************************************************************/
    /***   At how many multipoles are you evaluating the biased ang PS (Tl)? ***/
    /***************************************************************************/    
    L_BINS  = file_eof_linecount("multipoles_to_evaluate_input.dat");
    printf("You are using %d multipoles at which to evaluate Tl/Cl (from file multipoles_to_evaluate_input.dat)\n", L_BINS);

    /**********************************/
    /** number of tomographic z-bins **/
    /**********************************/
    if (ALL_FILE_INPUT == 0)
    {
        printf("How many tomographic bins are you using?\n");

        if(scanf("%d%c", &Z_BINS, &term) != 2 || (term != '\n' && term != 32) )
        {
            printf("Error: number of tomographic bins needs to be an integer!\n");
            printf("Exiting...\n");
            exit(EXIT_FAILURE);
        }        
    }
    else
    {
        if(fscanf(ifp_all_file_input, "%d%c", &Z_BINS, &term) != 2 || (term != '\n' && term != 32) )
        {
            printf("Error: number of tomographic bins needs to be an integer!\n");
            printf("Exiting...\n");
            exit(EXIT_FAILURE);
        }     
        fgets(mystring, MAX_LENGTH, ifp_all_file_input);  // read the rest (comments) of that line
        //printf("rest of 1st line: %s\n", mystring);
    }

    if (Z_BINS != count_words_in_row_of_file("Cl_fiduc_input.dat")-1)
    {
        printf("ERROR, number of power spectra in Cl_fiduc_input.dat has to match number of tomographic bins\n");
        printf("Exiting...\n");
        exit(EXIT_FAILURE); 
    }            

    
    /********************************************************************************/
    /***   pairs of tomographic bins; only i<=j, but in this code, for now                ***/
    /***   USE ONLY THE AUTO-CORRELATIONS i=j (so alpha=ij_to_alpha(i, i) further below)  ***/
    /********************************************************************************/

    Z_BIN_PAIRS = ij_to_alpha(Z_BINS, Z_BINS);  

    /******************************/
    /***   Define basic arrays  ***/
    /******************************/
    Cl_calib = dvector(1, L_MAX_CALIB);

    mag_to_calib_fac = dvector(1, Z_BINS);
    sz_arr           = dvector(1, Z_BINS);

    l2_arr  = imatrix(1, L_BINS, 1, 2*L_MAX_CALIB+1);
    l2_1D_arr  = ivector(1, (2*L_MAX_CALIB+1)*L_BINS);  // note number of entries

    mag_to_calib_fac = dvector(1, Z_BINS);
    sz_arr           = dvector(1, Z_BINS);
    
    Cl_matrix         = dmatrix(1, Z_BIN_PAIRS, 1, L_BINS);
    Tl_matrix         = dmatrix(1, Z_BIN_PAIRS, 1, L_BINS);

    /***************************************************************************************/
    /** read the fiducial power spectrum Cl at given successive multipoles (say l=1-1000) **/
    /***************************************************************************************/

    L_MAX  = file_eof_linecount("Cl_fiduc_input.dat");
    printf("Total number of multipoles of your power spectrum is %d (from file Cl_fiduc_input.dat)\n", L_MAX);

    Cl_all_ell =  dmatrix(1, Z_BIN_PAIRS, 1, L_MAX+L_MAX_CALIB); // to account for coupling even at L_MAX    

    ifp = fopen("Cl_fiduc_input.dat", "r");
    for(i=1; i<=L_MAX; i++)
    {
        fscanf(ifp, "%d ", &l);
        for(iz=1; iz<=Z_BINS; iz++) // ignoring the cross-bin power spectra for now
        {
            alpha = ij_to_alpha(iz, iz); 
            fscanf(ifp, "%lf ", &Cl_all_ell[alpha][i]);
        }
        fscanf(ifp, "\n" );
        
        if (i != l)
        {
            printf("Read-in ERROR in your Cl_fiduc_input.dat file\n");
            printf("Exiting...\n");
            exit(EXIT_FAILURE);
        }
    }
    fclose(ifp);

    /*******************************************/
    /***   Define the 3D array of clm coeff  ***/
    /*******************************************/
    c_arr = d3tensor(1, lround(L_MAX)+2, 1, lround(L_MAX)+3, 1, 2);

    for (l1 = 0; l1<lround(L_MAX)+2; l1++)
        for (m1 = 0; m1<lround(L_MAX)+3; m1++)
            c_arr[l1+1][m1+1][1] = c_arr[l1+1][m1+1][2] = 0;

    /***************************************************/
    /*** prompt user for input - dN/N vs. magnitudes ***/
    /***************************************************/
    
    if (ALL_FILE_INPUT == 0) // if input from the keyboard
    {
        printf("Would you like to input MAGNITUDE errors (1) or dN/N errors (2)?\n");

        if(scanf("%d%c", &CAL_OR_MAG, &term) != 2 || (term != '\n' && term != 32) )
        {
            printf("Error: This choice (CAL_OR_MAG) needs to be an integer!\n");
            printf("Exiting...\n");
            exit(EXIT_FAILURE);
        }
    
        if (CAL_OR_MAG == 1)
        {
            printf("You need to specify the faint-end slope of LF to convert from mag to dN/N\n");
            printf("Input the LF slope s(z) in each redshift bin (need real numbers):\n");
            for (iz=1; iz<=Z_BINS; iz++)
                scanf("%lf", &sz_arr[iz]);
        }
    }
    else // if input from the file
    {
        if(fscanf(ifp_all_file_input, "%d%c", &CAL_OR_MAG, &term) != 2 || (term != '\n' && term != 32) )
        {            
            printf("Error: CAL_OR_MAG choice needs to be an integer!\n");
            printf("Exiting...\n");
            exit(EXIT_FAILURE);
        }
        fgets(mystring, MAX_LENGTH, ifp_all_file_input);  // read the rest (comments) of that line
        // printf("rest of CAL_OR_MAG line: %s\n", mystring);

        if (CAL_OR_MAG == 1)        // if mag, read the sz in each tomographic bin
        {
            for (iz=1; iz<=Z_BINS; iz++)
                fscanf(ifp_all_file_input, "%lf ", &sz_arr[iz]);
            fscanf(ifp_all_file_input, "\n");

            fgets(mystring, MAX_LENGTH, ifp_all_file_input);  // read the rest (comments) of that line
            // printf("rest of s(z) values line %s\n", mystring);
        }
        if (CAL_OR_MAG == 2)        // if dN/N, skip this line in file
        {
            fgets(mystring, MAX_LENGTH, ifp_all_file_input);
            // printf("ALL of s(z) values line %s\n", mystring);
        }
    }
    
    if ((CAL_OR_MAG != 1) && (CAL_OR_MAG != 2))
    {
        printf("Error: Your CAL_OR_MAG input has to be 1 or two!\n");
        printf("Exiting...\n");
        exit(EXIT_FAILURE);
    }

    
    for (iz=1; iz<=Z_BINS; iz++)
    {
        if (CAL_OR_MAG == 1)       // fixed calibration rms
            mag_to_calib_fac[iz] = log(10.0) * sz_arr[iz];
        else                       // fixed magnitude var - conversion to calib is ln(10)*s(z)
            mag_to_calib_fac[iz] = 1.0;
    }
    
    /******************************************************************************/
    /*** prompt user for input - rms over some l-range vs. give power of errors ***/
    /******************************************************************************/
    if (ALL_FILE_INPUT == 0) // if input from the keyboard
    {
        printf("Would you like to:\n");
        printf("Just provide the rms of magnitude (or dN/N) variations                           (1)\n");
        printf("Input the angular PS of calibration (mag or dN/N) field in file Cl_mag_input.dat (2)\n");

        if(scanf("%d%c", &CHOICE, &term) != 2 || (term != '\n' && term != 32) )
        {
            printf("Error: This choice (CHOICE variable) needs to be an integer!\n");
            printf("Exiting...\n");
            exit(EXIT_FAILURE);
        }

        if ((CHOICE != 1) && (CHOICE != 2))
        {
            printf("Error: Your CHOICE input has to be 1 or 2!\n");
            printf("Exiting...\n");
            exit(EXIT_FAILURE);
        }

        if (CHOICE == 1)
        {
            printf("How many multipoles (ie. what LMAX) would you like to spread this (rms)^2 over? (e.g. 10)\n");
            if(scanf("%d%c", &LMAX_SPREAD_CVAR, &term) != 2 || (term != '\n' && term != 32) )
            {
                printf("Error: This choice (LMAX_SPREAD_VAR) needs to be an integer!\n");
                printf("Exiting...\n");
                exit(EXIT_FAILURE);
            }            
            
            printf("What total rms of error would you like over this l-range (e.g. 0.01)?\n");
            scanf("%lf", &CALIB_OR_MAG_ERR);
            
            // assign the c_arr consistent with rms=CALIB_OR_MAG_ERR
            assign_clm_given_fixed_rms(CALIB_OR_MAG_ERR, LMAX_SPREAD_CVAR, c_arr);
            
        }
    }
    else // if input from the file
    {
        if(fscanf(ifp_all_file_input,"%d%c", &CHOICE, &term) != 2 || (term != '\n' && term != 32) )
        {
            printf("Error: The CHOICE variable needs to be an integer! %d %d\n", CHOICE, term);
            printf("Exiting...\n");
            exit(EXIT_FAILURE);
        }
        fgets(mystring, MAX_LENGTH, ifp_all_file_input);  // read the rest (comments) of that line
        //printf("Rest of CHOICE input line: %s\n", mystring);

        if ((CHOICE != 1) && (CHOICE != 2))
        {
            printf("Error: The CHOICE input has to be 1 or 2!\n");
            printf("Exiting...\n");
            exit(EXIT_FAILURE);
        }
        
        if (CHOICE == 1) // read the two distribution and size of rms
        {
            fscanf(ifp_all_file_input, "%d ", &LMAX_SPREAD_CVAR);
            //printf("LMAX_SREAD_CVAR=%d\n", LMAX_SPREAD_CVAR);
            fgets(mystring, MAX_LENGTH, ifp_all_file_input);  // read the rest (comments) of that line

            fscanf(ifp_all_file_input, "%lf ", &CALIB_OR_MAG_ERR);            
            //printf("CALIB_OR_MAG_ERR=%f\n", CALIB_OR_MAG_ERR);
            fgets(mystring, MAX_LENGTH, ifp_all_file_input);  // read the rest (comments) of that line

            // assign the c_arr consistent with rms=CALIB_OR_MAG_ERR
            assign_clm_given_fixed_rms(CALIB_OR_MAG_ERR, LMAX_SPREAD_CVAR, c_arr);
        
        }
        else // just skip those two lines in the file
        {
            fgets(mystring, MAX_LENGTH, ifp_all_file_input);
            fgets(mystring, MAX_LENGTH, ifp_all_file_input);
        }
    }
    

    if (CHOICE == 2)
    {
        printf("You should have the file Cl_mag_input.dat in this directory\n");
        printf(" It should have L_CALIB_MAX rows in two-column format: l Cl\n");

        
        ifp=fopen("Cl_mag_input.dat", "r");
        if (ifp == NULL) {
            fprintf(stderr, "Error: failed to open file %s\n", filename);
            printf("Exiting...\n");            
            exit(EXIT_FAILURE);
        }
        for(i=1; i<=L_MAX_CALIB; i++)
        {
            fscanf(ifp, "%d %lf\n", &l, &Cl_calib[i]);
            
            if (i != l)
            {
                printf("ERROR, your Cl_mag_input.dat array has to start with l=1\n");
                printf("i=%d l=%d\n", i, l);
                printf("Exiting...\n");
                exit(EXIT_FAILURE); 
            }
        }
        fclose(ifp);        
        assign_clm_given_Cl_calib(Cl_calib, c_arr);
    }

    if (ALL_FILE_INPUT == 1)
        fclose(ifp_all_file_input );
    
    /******************************************************************************************/
    /**        read the list of multipoles at which to evaluate biased angPS Tl              **/
    /**   and assign the Cl_matrix values to be Cl_all_ell at those particular l             **/
    /** note: ij_to_alpha(iz, jz) converts from two indices for a bin (iz, jz) to one (alpha)**/
    /******************************************************************************************/      
    L_BINS  = file_eof_linecount("multipoles_to_evaluate_input.dat");
    l_arr  = ivector(1, L_BINS); // array of mappings from l-bin to l

    ifp=fopen("multipoles_to_evaluate_input.dat", "r");
    for(il=1; il<=L_BINS; il++)
        fscanf(ifp, "%d\n", &l_arr[il]);
    fclose(ifp);

    for (iz=1; iz<=Z_BINS; iz++)
        for(jz=iz; jz<=Z_BINS; jz++)                
            for(il=1; il<=L_BINS; il++)
                Cl_matrix[ij_to_alpha(iz, jz)][il] =  Cl_all_ell[ij_to_alpha(iz, jz)][l_arr[il]];
            
    

    /***************************************************************************************/
    /**                                 get the ell central values                        **/
    /** also get all of the values of l2, which range from l-L_MAX_CALIB to l+L_MAX_CALIB **/
    /***************************************************************************************/
   
    i=1;
    for(il=1; il<=L_BINS; il++)
    {            
        /********************************************************************/
        /**         create the l2 array, and also its 1-D version          **/
        /* Checked explicitly that l2_arr and l2_1D_arr agree (13 Nov 2011) */
        /********************************************************************/
        for(il2=1; il2<=(2*L_MAX_CALIB+1); il2++)
        {
            l2_arr[il][il2] = l_arr[il]-L_MAX_CALIB+(il2-1); // [l-L_MAX_CALIB,l+L_MAX_CALIB]
            l2_1D_arr[i] = l2_arr[il][il2];
            i++;
        }
    }
    N_L2 = i-1;
    if (N_L2 != (2*L_MAX_CALIB+1)*L_BINS)
    {
        printf("Something wrong here! N_L2=%d (2*L_MAX_CALIB+1)*L_BINS=%d\n", N_L2 , (2*L_MAX_CALIB+1)*L_BINS);
        printf("Exiting...\n");
        exit(EXIT_FAILURE);
    } 

    
    /***********************************************************************************/
    /* need to call get_alm before you run calculate_Tl or call epsilon the first time */
    /***********************************************************************************/
    alm_RE = d3tensor(1, Z_BINS, 1, L_MAX_CALIB+1, 1, L_MAX_CALIB+1);  // index=1 => l=0 etc
    alm_IM = d3tensor(1, Z_BINS, 1, L_MAX_CALIB+1, 1, L_MAX_CALIB+1);  // index=1 => l=0 etc

    for (l1 = 1; l1 <= L_MAX_CALIB+1; l1++)
        for (m1 = 1; m1 <= L_MAX_CALIB+1; m1++)
            for(iz=1; iz<= Z_BINS; iz++)
                alm_RE[iz][l1][m1] = alm_IM[iz][l1][m1] =0; // just in case
 
    get_alm(Cl_all_ell);  // get the (random) alm for galaxy Cl - needed for epsilon term(s)

    /**************************************************************/
    /**   Calculate the Tl/Cl biases due to calibration error    **/
    /**************************************************************/
    calculate_Tl(c_arr, Cl_all_ell, mag_to_calib_fac, Cl_matrix, Tl_matrix);

    /**********************************/
    /** Report how much time it took **/
    /**********************************/
    double t=omp_get_wtime();    
    printf("Finished!  Time to  calculate_Tl= %lf seconds\n", omp_get_wtime()-t);              


  exit(1);
  }

