/*******************************************************************************************/
/*** Principal routines that calculate the Wigner-3j symbols and other geometrical terms ***/
/*** and calculate the biased angular power spectra, Tl                                  ***/
/*******************************************************************************************/

#include<omp.h>
#include<math.h>
#include<stdio.h>
#include <gsl/gsl_sf.h>
#include"nrutil.h" 
#include"nr.h"
#include"declare.h"
#include"wigner.h" // crucial for R at high ell!
 
void calculate_Tl(double ***c_array, double **Cl_all_ell, double *mag_to_calib_fac, double **Cl, double **Tl)
{
    /*********************************************************************************/
    /** Calculates the biased Cl[alpha][il], called Tl[alpha][il]                   **/
    /** Used only in specific maps now (PG10, Annis), and not for cal at each ell   **/
    /** note, it still refers to *real* multipoles (e.g. l, not il) in 2nd argument **/
    /** note, Cl        [alpha][il] is defined at binned ell, while                 **/
    /**       Cl_all_ell[iz][l]  is defined at ALL ell                              **/
    /*********************************************************************************/
    int l, m, l2, m2, iz, jz, alpha, index1, index2;
    int il, il2;
    double *sum_2, U_term_ave_over_m, Usq_term, eps,eps_i, eps_j, sum_3;
    double magtocalib_quadratic_U_term, magtocalib_linear_U_term;
    FILE *ifp;

    sum_2 = dvector(1, Z_BIN_PAIRS);

    ifp =fopen("Tl_over_Cl_output.dat", "w"); 

    printf("*******************************************************************************\n");
    printf("Printing out 1st z-bin on screen (all z-bins are in Tl_over_Cl_output.dat file)\n");
    printf("(Also: printing two multiplicative and one additive contribution to that Tl/Cl)\n");
    printf("*******************************************************************************\n");
    printf("   l   Tl/Cl    (1+eps)^{-2}   Mult-1   Mult-2   Additive\n");
    for(il=1; il<=L_BINS; il++)  // go over all binned values of l
    {
        l = l_arr[il]; 
        
        generate_U_Usq_single_l(il, c_array);   // generate U!

        for(alpha=1; alpha<=Z_BIN_PAIRS;  alpha++)
            sum_2[alpha] = sum_3 = 0;

        /**************************************/
        /* this is just for l1=0, basically! **/
        /**************************************/
        U_term_ave_over_m = 2 * c_array[1][1][1] / sqrt(4.0*M_PI);
        
        for(m=-l; m<=l; m++)
        {
            /*********************************************/
            /* add the |clm|^2 term, sum it over m first */
            /*********************************************/
            if (m == 0)                        
                sum_3 += pow(c_array[l+1][m+1][1], 2) ;
            else                        
                sum_3 += pow(c_array[l+1][m+1][1], 2) + pow(c_array[l+1][m+1][2], 2);  // no 2 since m-sum starts at -l 
            
            // go over l2=[l-L_MAX_CALIB, l+L_MAX_CALIB]; do only of l2 >= 0
            for (il2=1; il2<=2*L_MAX_CALIB+1; il2++)
            {
                l2 = l2_arr[il][il2]; // integer
                if (l2 >= 0)
                {
                    for(m2=-l2; m2<=l2; m2++)
                    {
                        index1 = justm_2index_compressed(il, m);    
                        index2 = l2m2_2index_compressed(il,il2, m2);   


                        Usq_term = Usq_l_m_l2_m2[index1][index2];
                        
                        for(alpha=1; alpha<=Z_BIN_PAIRS; alpha++)
                            sum_2[alpha] += Usq_term * Cl_all_ell[alpha][l2];
                    } // end of sum over m2
                }
                
            } // end of sum over l2
        } // end of sum over m

        /******************************************************************/
        /** Free the U matrices - CRUCIAL TO SPEED UP memory allocation! **/
        /******************************************************************/
        //free_d3tensor(U_l_m_l2_m2,   1, SIZE_INDEX1, 1, SIZE_INDEX2, 1, 2);
        free_dmatrix (Usq_l_m_l2_m2, 1, SIZE_INDEX1, 1, SIZE_INDEX2);
        
        
        /***************************************************************************/
        /** Calculate the biased Cl (Tl) and  average the corrections over m in sum2*/
        /*****************************************************************************/
        for(iz=1; iz<=Z_BINS; iz++)
            for(jz=iz; jz<=Z_BINS; jz++)
            {            
                //eps = epsilon(iz, c_array, Cl_all_ell);
                //Tl[iz][il]  = pow(1+eps, -2) * (Cl[iz][il]*(1+U_term_ave_over_m) + (sum_2[iz] + sum_3) / (2.0*l+1) );
                
                eps_i = epsilon(iz, c_array) * mag_to_calib_fac[iz];
                eps_j = epsilon(jz, c_array) * mag_to_calib_fac[jz];

                magtocalib_linear_U_term    = 0.5*(mag_to_calib_fac[iz] + mag_to_calib_fac[jz]);
                magtocalib_quadratic_U_term = mag_to_calib_fac[iz] * mag_to_calib_fac[jz];
                
                alpha = ij_to_alpha(iz, jz);
                
                Tl[alpha][il]  = 1.0/((1+eps_i)*(1+eps_j)) *
                    (Cl[alpha][il]*(1+U_term_ave_over_m * magtocalib_linear_U_term) +
                     magtocalib_quadratic_U_term * (sum_2[alpha] + sum_3) / (2.0*l+1) );
            }

        /***********************************************************************************/
        /*** To print Tl/Cl, for auto-correlations only (so iz=jz, same tomographic bin)  **/
        /*** on screen, print only for 1st bin, but in file print all iz ratios           **/
        /***********************************************************************************/
        iz=1; // really alpha; any value will do for printing
        alpha = ij_to_alpha(iz, iz);
        eps = epsilon(iz, c_array);        
        fprintf(ifp, "%d %f ", l, Tl[alpha][il] / Cl[alpha][il]);

        printf("%4d %f  (eps %f) %f %f %f\n", l, Tl[alpha][il]/Cl[alpha][il], pow(1+eps, -2)*(1+U_term_ave_over_m), 
               U_term_ave_over_m, sum_2[alpha]/(2.0*l+1)/Cl[alpha][il], sum_3/(2.0*l+1)/Cl[alpha][il]);
        for(iz=2; iz<=Z_BINS; iz++)
        {
            alpha = ij_to_alpha(iz, iz);
            fprintf(ifp, "%f ", Tl[alpha][il] / Cl[alpha][il]);
        }
        fprintf(ifp, "\n");

    } // end of the while loop over il (binned multipoles)
    fclose(ifp);
    printf("*******************************************************************************\n");
    printf("Written the file Tl_over_Cl_output.dat\n");

}
int justm_2index_compressed(int il, int m)
{
    /*******************************************************/
    /**   same as lm_2index_compressed, but l is fixed    **/
    /**              Access as l+m+1                      **/
    /*******************************************************/

    return (lround(l_arr[il]) + m+1);
}
int l2m2_2index_compressed(int il, int il2, int m2)
{
    /*******************************************************************/
    /**                     refurbished on Feb 12, 2012.              **/
    /**  Uses only the necessary range in (l2, m2) which depends on l **/
    /**  The maximum size is comparable to (2*L_MAX_CALIB+1)*(2*l+1)  **/
    /**     offset takes into account maximal negative departure      **/
    /**                 in cases where L_MAX_CALIB > l                **/
    /*******************************************************************/
    int offset, index2;
    int l2 = l2_arr[il][il2];
    int l  = l_arr[il];

    if (L_MAX_CALIB > l) offset = (L_MAX_CALIB-l)*(L_MAX_CALIB-l);
    else                 offset = 0;

    index2 = (il2-1) * (2*(l-L_MAX_CALIB)+il2-1) + l2 + m2 + 1 + offset;

    if (index2 < 0)
    {
        printf("WARNING! - index2 < 0\n");
        exit(1);
    }
    
    return (index2);
}
double R(int l1, int l2, int l3, int m1, int m2, int m3)
{
    /***************************************************************************/
    /** returns the Wigner 3j function, or zero if one of the l's is negative **/
    /********************************************************************************/
    /** REMEMBER THAT gsl_sf_coupling_3j TAKES ARGUMENTS IN HALF_INTEGER UNITS!!!! **/
    /********************************************************************************/
    double result;
    
    //if (m1+m2-m3 != 0)
        //printf("The m entries don't sum to zero %d %d %d\n", m1, m2, m3);

    if ((l1 < 0) || (l2 < 0) || (l3 < 0))
        result = 0;
    else if (l3 < MIN_ELL_USE_APPROX) // use exact Wigner 3j
        result = pow(-1.0, m3) * sqrt( (2*l1+1.0) * (2*l2+1.0) * (2*l3+1.0) / (4*M_PI))
            * gsl_sf_coupling_3j(2*l1, 2*l2, 2*l3, 2*0, 2*0, 2*0)
            * gsl_sf_coupling_3j(2*l1, 2*l2, 2*l3, 2*m1,2*m2, 2*(-m3));
    else  // use approximate Wigner 3j that works only for l1<<l2, l3
        result = pow(-1.0, m3) * sqrt( (2*l1+1.0) * (2*l2+1.0) * (2*l3+1.0) / (4*M_PI))
            * gsl_sf_wigner_3j(2*l1, 2*l2, 2*l3, 2*0, 2*0, 2*0)
            * gsl_sf_wigner_3j(2*l1, 2*l2, 2*l3, 2*m1,2*m2, 2*(-m3));
    return result;
}

void generate_U_Usq_single_l(int il, double ***c_array)
/***********************************************************************************/
/** You need to save U only once at a given l (as a function of m, l2, m2)        **/
/***********************************************************************************/
{
    int l1, l2, l, m, m1, m2, index1, index2;
    int  il2;
    double sum_imag, sum_real, dsum_imag, dsum_real, m1neg;

    /***************************************************************/
    /** get the l, since that also determines the index1 size     **/
    /** SIZE_INDEX1 defined here; SIZE_INDEX2 defined in Cl_LSS.c **/
    /***************************************************************/
    l = (int) lround(l_arr[il]);
    SIZE_INDEX1 = 2*l+1;  // number of m for that l
    SIZE_INDEX2 = 4 * lround(L_MAX)*L_MAX_CALIB + 2*(lround(L_MAX) + L_MAX_CALIB) + 1;
      
    /************************************/
    /** define the U arrays every time **/
    /************************************/
    //U_l_m_l2_m2     = d3tensor(1, SIZE_INDEX1, 1, SIZE_INDEX2, 1, 2);
    Usq_l_m_l2_m2   = dmatrix (1, SIZE_INDEX1, 1, SIZE_INDEX2);

    for(index1=1; index1<=SIZE_INDEX1; index1++)
        for(index2=1; index2<=SIZE_INDEX2; index2++)
            Usq_l_m_l2_m2[index1][index2] = 0;

    /****************************************************************************************/
    /**  PARALLEL OPENMP command; the only variable NOT private is outermost m, and also l **/
    /****************************************************************************************/    
    
#pragma omp parallel for private(l2, m2, l1, m1, m1neg, il2, sum_real, sum_imag, dsum_real, dsum_imag, index1, index2) shared(c_array, Usq_l_m_l2_m2)  schedule(dynamic)

    for(m=-l; m<=l; m++)
    {
        //int th_id = omp_get_thread_num();
        //printf("m =%d from thread  %d\n", m , th_id);

        for(il2=1; il2<=2*L_MAX_CALIB+1; il2++)
        {
            l2 = l2_arr[il][il2];  // correct range; I checked 29 Jan 2012
            //printf("il=%d l2=%d\n", il, l2);

            if (l2 >= 0)  // need this at low-ish l when l2 can go unphysically <0
                for(m2=-l2; m2<=l2; m2++)
                {
                    sum_real = sum_imag = 0;
                    
                    for(l1=0; l1<=L_MAX_CALIB; l1++) // going to lower ell - this is contam
                    {
                        if (condition(l2,l1,l, m2)) // if condition fails to hold
                            dsum_real=dsum_imag=0;
                        else                         
                            for(m1=0; m1<=l1; m1++) // start at zero
                            {
                                dsum_real=dsum_imag=0;                            
                                
                                if ((m1 == m-m2) || (-m1 == m-m2)) // cut work shorter by waiting for m1-m2-m (or -m1) to be 'right'
                                {
                                    dsum_real += c_array[l1+1][m1+1][1] * R(l1, l2, l, m1, m2, m);
                                    dsum_imag += c_array[l1+1][m1+1][2] * R(l1, l2, l, m1, m2, m);

                                    if (m1 != 0)
                                    {
                                        m1neg = -m1; // flip m1, but leave m2 and wait for it to be 'right'

                                        dsum_real +=   c_array[l1+1][m1+1][1] * R(l1, l2, l, m1neg, m2, m);
                                        dsum_imag += - c_array[l1+1][m1+1][2] * R(l1, l2, l, m1neg, m2, m);
                                    }
                                }
                                sum_real += dsum_real;
                                sum_imag += dsum_imag;
                                
                            } // ends m1 loop
                        
                    } // ends l1 loop
                    
                    /* now that you are finished with this (l1, m2), get U(l, m, l2, m2) basically */
                    index1 = justm_2index_compressed(il, m) ;
                    index2 = l2m2_2index_compressed (il, il2, m2);
                    
                    Usq_l_m_l2_m2[index1][index2] = pow(sum_real, 2) + pow(sum_imag, 2) ;                    
                    
                } // closes the m2 loop
        } // closes the il2 loop
        
    } // closes the m loop

    // PARALLEL OPENMP command
    #pragma omp barrier    // doesn't appear necessary
    
}
double epsilon(int iz, double ***c_arr)
{
    int l, m;
    double sum=c_arr[1][1][1]/sqrt(4*M_PI); // first term


    /************************************************/
    /* get a set of alm given the power spectrum Cl */
    /************************************************/
    
    for (l = 0; l <= L_MAX_CALIB; l++)
    {
        sum += 1.0/(4*M_PI) * c_arr[l+1][1][1]*alm_RE[iz][l+1][1];  // c_l0*a_l0/(4*PI)

        for (m = 1; m <= l; m++)                 // l1, m1>0
            sum += 2/(4*M_PI)*(c_arr[l+1][m+1][1]*alm_RE[iz][l+1][m+1] +
                               c_arr[l+1][m+1][2]*alm_IM[iz][l+1][m+1]); // c_lm*a_lm^*/(4*PI)
    }

    //printf("Epsilon for bin %d is %f\n", iz, sum);
    
    return(sum);
}
int condition(int l2, int l1, int l, int m2)
{
    // returns one if (l, l1, l2) do NOT satisfy the triangle condition
    // also returns one if m2 is WRONG (out of bounds, which can happen when m2=m-m1 is assigned (in fixed-m1 cases)
    // returns zero otherwise

    return ((l > l1+l2) || (l2 > l+l1) || (l1 > l+l2) ||
            (l < abs(l1-l2)) || (l2 < abs(l-l1)) || (l1 < abs(l-l2)) || (abs(m2) > l2));
}

